import json
import logging
import os
from datetime import datetime

from microwler.settings import Settings

LOG = logging.getLogger(__name__)


class BaseExporter:
    """
    Use this class to build your custom export functionality, i.e. send data per HTTP or SMTP.
    The crawler instance will call `export()` once it's done with everything else.
    You can pass your plugin into the crawler by adding the class to `settings['exporters']`
    """

    def __init__(self, domain: str, data: list, settings: Settings):
        """
        Create a new BaseExporter

        Arguments:
            domain: the domain of this project/crawler
            data: list of processed Page objects
            settings: the current settings of this project/crawler
        """
        self.domain = domain
        self.data = [page.__dict__ for page in data]
        self.settings = settings

    def export(self):
        """
        Export data to target destination
        """
        raise NotImplementedError()


class FileExporter(BaseExporter):
    """
    This exporter will save data to your local filesystem. It currently provides
    exports to JSON, CSV or HTML tables. Take a look at the following exporters
    and their implementation to understand its usage.
    """
    extension = ''

    def convert(self):
        """
        Converts `self.data` to output format specified by `FileExporter.extension`.
        > Must return converted data as `string`
        """
        raise NotImplementedError()

    def export(self):
        """ Writes data to file """
        data = self.convert()
        timestamp = datetime.now().strftime('%Y-%m-%d-%H:%M')
        path = os.path.join(self.settings.export_to, f'{self.domain}_{timestamp}.{self.extension}')
        try:
            os.makedirs(self.settings.export_to, exist_ok=True)
            with open(path, 'w') as file:
                file.write(data)
            LOG.info(f'Exported data as {self.extension.upper()} to: {path} [{self.domain}]')
        except Exception as e:
            LOG.error(f'Error during export: {e} [{self.domain}]')


class JSONExporter(FileExporter):
    """ Exports to JSON files """
    extension = 'json'

    def convert(self):
        data = json.dumps(self.data)
        return data


class CSVExporter(FileExporter):
    """ Exports to CSV files """
    extension = 'csv'

    def convert(self):
        headers = ';'.join([key.upper() for key in self.data[0].keys()])
        rows = '\n'.join([';'.join([str(val) for val in obj.values()]) for obj in self.data])
        table = '\n'.join([headers, rows])
        return table


class HTMLExporter(FileExporter):
    """ Exports data as <table> to HTML files """
    extension = 'html'

    def convert(self):
        styles = 'width: 100%; border: 1px solid grey; text-align: center'
        headers = ''.join([f'<th>{key.upper()}</th>' for key in self.data[0].keys()])
        rows = ''.join([f"<tr>{''.join([f'<td>{val}</td>' for val in obj.values()])}</tr>" for obj in self.data])
        table = f'<!DOCTYPE html><html><body><table style="{styles}"><tr>{headers}</tr><tbody>{rows}</tbody></table><body></html>'
        return table






TEMPLATE = """# This file was auto-generated by Microwler

# Use this template as starting point for your new project.
# To find out more about configuration options, read the docs: https://innovinati.github.io/microwler


from microwler import Microwler, scrape, export


select = {
    # Add selectors here
    # Docs: https://innovinati.github.io/microwler/scraping/#selectors
}

settings = {
    # Add settings here
    # Docs: https://innovinati.github.io/microwler/configuration/#settings
}


def transform(data: dict):
    # Manipulate data here
    # Example: https://github.com/INNOVINATI/microwler/blob/master/test_cases.py#L54
    return data


# Microwler expects a variable "crawler" when running from CLI
crawler = Microwler(
    'START_URL',
    select=select,
    transform=transform,
    settings=settings
)
"""



import datetime
import logging

from lxml.etree import ParserError
from parsel import Selector

from microwler.utils import get_first_or_list


LOG = logging.getLogger(__name__)


class Page(object):
    """
    Internal representation of a webpage

    > The methods `scrape()` and `transform()` will be called by the crawler instance
    > if selectors and a corresponding transformer function are defined.
    """

    def __init__(self, url: str, status_code: int, depth: int, links: list = None, html: bytes = None):
        """
        Arguments:
            url: the URL of this page
            status_code: the request's getStatus code
            depth: the depth at which this page was crawled
            links: list of internal links found on this page
            html: HTML body
        """
        self.url = url
        self.discovered = datetime.date.today().strftime('%Y-%m-%d')
        self.status_code = status_code
        self.depth = depth
        self.links = links
        self.html = html
        self.data = {}

    def scrape(self, selectors: dict, keep_source=False):
        """
        Extracts data using the given selectors. Selectors are either `XPaths` (strings) or callables.
        If a callable is given, it will receive the parsed DOM as only argument,
        which is a [Parsel selector](https://parsel.readthedocs.io/en/latest/usage.html#using-selectors) instance.
        This means you can apply `dom.xpath(...)` or `dom.css(...)` and return any native Python datatype you want.

        > Note: The selected data items will be stored in `Page.data`, which is a Python `dict`
        """

        try:
            dom = Selector(text=self.html)
            for field, selector in selectors.items():
                if type(selector) == str:
                    self.data[field] = get_first_or_list(dom.xpath(selector))
                else:
                    self.data[field] = selector(dom)
        except ParserError as e:
            LOG.warning(f'Parsing error: {e}')

        if not keep_source:
            del self.html

        return self

    def transform(self, func):
        """
        Applies a given function to this page's data.
        Recommended usage of transformers is to manipulate the input (`self.data`)
        and return the whole thing. If nothing is returned, the page's data property
        will be set to None.

        Arguments:
            func: a Python callable
        """
        if len(self.data):
            try:
                self.data = func(self.data)
                return self
            except Exception as e:
                LOG.warning(f'Transformer error: {e}')
                return self
        raise ValueError('You need to provide selectors in order to use a transformer')



import logging
import os

LOG = logging.getLogger(__name__)


class Settings(object):
    link_filter: str = '//a/@href'
    max_depth: int = 10
    max_concurrency: int = 20
    dns_providers: list = ['1.1.1.1', '8.8.8.8']
    language: str = 'en-us'
    caching: bool = False
    delta_crawl: bool = False
    export_to = os.path.join(os.getcwd(), 'exports')
    exporters: list = []

    def __init__(self, params: dict):
        if params:
            for key, value in params.items():
                setattr(self, key, value)

            if self.delta_crawl and not self.caching:
                self.caching = True
                LOG.info('Auto-enabled caching (required for delta_crawl)')



import os
import sys
from urllib.parse import urlparse

import click
from click.testing import CliRunner

from microwler.cli.template import TEMPLATE
from microwler.utils import load_project, PROJECT_FOLDER
from microwler.web.backend import start_app

HERE = os.path.dirname(os.path.abspath(__file__))
COMMANDS = [
    ('new PROJECT_NAME START_URL', 'Create a new project'),
    ('crawler PROJECT_NAME run', 'Run a project\'s crawler'),
    ('crawler PROJECT_NAME dumpcache', 'Dump project getCache to JSON file'),
    ('crawler PROJECT_NAME clearcache', 'Clear project getCache'),
    ('serve [-p|--port]', 'Start the built-in webservice'),
]


@click.command('microwler')
def show_help():
    """ Show available commands """
    click.echo(
        """
╔╦╗┬┌─┐┬─┐┌─┐┬ ┬┬  ┌─┐┬─┐
║║║││  ├┬┘│ │││││  ├┤ ├┬┘
╩ ╩┴└─┘┴└─└─┘└┴┘┴─┘└─┘┴└─            
- powered by INNOVINATI -                                                                              
        """
    )
    print('USAGE:')
    for cmd, description in COMMANDS:
        print(f'> {description}:\n  {cmd}\n')


@click.command()
@click.argument('project_name', type=str)
@click.argument('start_url', type=str)
def add_project(project_name, start_url):
    """ Create a new project """
    try:
        os.makedirs(PROJECT_FOLDER, exist_ok=True)
        template = TEMPLATE.replace('START_URL', start_url)
        name = project_name.replace(' ', '_').replace('.', '_').replace('-', '_').lower()
        path = os.path.join(PROJECT_FOLDER, name + '.py')
        if os.path.exists(path):
            question = click.style('A project with this name already exists. Do you want to overwrite it?', fg='yellow')
            if not click.confirm(question):
                exit(0)
        with open(path, 'w') as output:
            output.write(template)
        click.secho(f'Created new project: {path}', fg='green')
    except Exception as e:
        click.secho(f'Failed to create project: {e}', fg='red')
        exit(1)


@click.group()
@click.argument('project_name')
@click.pass_context
def crawler(ctx, project_name):
    ctx.ensure_object(dict)
    project = project_name.split('.')[0] if project_name.endswith('.py') else project_name
    if project + '.py' not in os.listdir(PROJECT_FOLDER):
        click.secho(f'Project "{project}" does not exist', fg='red')
        exit(1)
    ctx.obj['project'] = project


@crawler.command('run')
@click.option('-v', '--verbose', default=False, is_flag=True)
@click.option('-s', '--sort', default=False, is_flag=True)
@click.option('--keep-html', default=False, is_flag=True)
@click.pass_context
def run_crawler(ctx, verbose, sort, keep_html):
    """ Run a project's crawler"""
    project = load_project(ctx.obj['project'])
    project.crawler.run(verbose=verbose, sort_urls=sort, keep_source=keep_html)


@crawler.command('dumpcache')
@click.option('-p', '--path')
@click.pass_context
def dump_cache(ctx, path):
    """ Dump the project getCache to a JSON file """
    project = load_project(ctx.obj['project'])
    if project.crawler._cache is not None:
        project.crawler.dump_cache(path)
    else:
        click.secho('Cache is disabled for this project', fg='yellow')


@crawler.command('clearcache')
@click.pass_context
def clear_cache(ctx):
    """ Clear the project getCache """
    project = load_project(ctx['project'])
    if project.crawler._cache is not None:
        project.crawler.clear_cache()
    else:
        click.secho('Cache is disabled for this project', fg='yellow')


@click.command('serve')
@click.option('-p', '--port', type=int, default=5000, help='The port to run the webservice on.')
def start_server(port):
    """ Start the built-in webservice """
    if not len(os.listdir(PROJECT_FOLDER)):
        click.secho('Running webservice with empty project folder', fg='yellow')
    start_app(port)



from html_text import extract_text
import parsel

from microwler.utils import remove_multi_whitespace


def title(dom: parsel.Selector):
    """ Extract `<title>` tag """
    return dom.xpath('string(//title[1])').get()


def headings(dom: parsel.Selector):
    """ Extract first 3 levels of heading tags: `<h1>`, `<h2>`, `<h3>` """
    return {
        'h1': remove_multi_whitespace(dom.xpath('string(//h1[1])').getall()),
        'h2': remove_multi_whitespace(dom.xpath('string(//h2[1])').getall()),
        'h3': remove_multi_whitespace(dom.xpath('string(//h3[1])').getall()),
    }


def paragraphs(dom: parsel.Selector):
    """ Extract `<p>` tags """
    return dom.xpath('string(//p[1])').getall()


def text(dom: parsel.Selector):
    """ Extract and clean text content """
    return extract_text(dom.xpath('//body').get())


def meta(dom: parsel.Selector):
    """ Extract `<meta>` tags """
    tags = dom.xpath('//meta')
    return {tag.get('name'): tag.attrib['content'] for tag in tags}


def canonicals(dom: parsel.Selector):
    """ Extract `<link rel='canonical'>` tags """
    return dom.xpath('//link[@rel="canonical"]/@href').getall()


def schemas(dom: parsel.Selector):
    """ Extract itemtype schemas """
    schema_links = dom.xpath('//*[@itemtype]/@itemtype').getall()
    return [link.split('/')[-1] for link in schema_links]


def emails(dom: parsel.Selector):
    """ Extract email addresses from `<a>` tags """
    hrefs = dom.xpath('//a[starts-with(@href, "mailto")]/@href').getall()
    return [href.strip('mailto:') for href in hrefs]


def images(dom: parsel.Selector):
    """ Extract URLs from `<img>` tags """
    return dom.xpath('//img/@src').getall()
